# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import numpy as np
from numpy.testing import assert_equal, assert_allclose
import matplotlib
import pytest

import MDAnalysis as mda
from MDAnalysisTests.datafiles import (
    GRO,
    XTC,
    TPR,
    DihedralArray,
    DihedralsArray,
    RamaArray,
    GLYRamaArray,
    JaninArray,
    LYSJaninArray,
    PDB_rama,
    PDB_janin,
)
import MDAnalysis.analysis.dihedrals
from MDAnalysis.analysis.dihedrals import Dihedral, Ramachandran, Janin


class TestDihedral(object):

    @pytest.fixture()
    def atomgroup(self):
        u = mda.Universe(GRO, XTC)
        ag = u.select_atoms(
            "(resid 4 and name N CA C) or (resid 5 and name N)"
        )
        return ag

    def test_dihedral(self, atomgroup, client_Dihedral):
        # client_Dihedral is defined in testsuite/analysis/conftest.py
        # among with other testing fixtures. During testing, it will
        # collect all possible backends and reasonable number of workers
        # for a given AnalysisBase subclass, and extend the tests
        # to run with all of them.

        dihedral = Dihedral([atomgroup]).run(**client_Dihedral)
        test_dihedral = np.load(DihedralArray)

        assert_allclose(
            dihedral.results.angles,
            test_dihedral,
            rtol=0,
            atol=1.5e-5,
            err_msg="error: dihedral angles should " "match test values",
        )

    def test_dihedral_single_frame(self, atomgroup, client_Dihedral):
        dihedral = Dihedral([atomgroup]).run(
            start=5, stop=6, **client_Dihedral
        )
        test_dihedral = [np.load(DihedralArray)[5]]

        assert_allclose(
            dihedral.results.angles,
            test_dihedral,
            rtol=0,
            atol=1.5e-5,
            err_msg="error: dihedral angles should " "match test vales",
        )

    def test_atomgroup_list(self, atomgroup, client_Dihedral):
        dihedral = Dihedral([atomgroup, atomgroup]).run(**client_Dihedral)
        test_dihedral = np.load(DihedralsArray)

        assert_allclose(
            dihedral.results.angles,
            test_dihedral,
            rtol=0,
            atol=1.5e-5,
            err_msg="error: dihedral angles should " "match test values",
        )

    def test_enough_atoms(self, atomgroup, client_Dihedral):
        with pytest.raises(ValueError):
            dihedral = Dihedral([atomgroup[:2]]).run(**client_Dihedral)

    def test_dihedral_attr_warning(self, atomgroup, client_Dihedral):
        dihedral = Dihedral([atomgroup]).run(stop=2, **client_Dihedral)

        wmsg = "The `angle` attribute was deprecated in MDAnalysis 2.0.0"
        with pytest.warns(DeprecationWarning, match=wmsg):
            assert_equal(dihedral.angles, dihedral.results.angles)


class TestRamachandran(object):

    @pytest.fixture()
    def universe(self):
        return mda.Universe(GRO, XTC)

    @pytest.fixture()
    def rama_ref_array(self):
        return np.load(RamaArray)

    def test_ramachandran(self, universe, rama_ref_array, client_Ramachandran):
        rama = Ramachandran(universe.select_atoms("protein")).run(
            **client_Ramachandran
        )

        assert_allclose(
            rama.results.angles,
            rama_ref_array,
            rtol=0,
            atol=1.5e-5,
            err_msg="error: dihedral angles should " "match test values",
        )

    def test_ramachandran_single_frame(
        self, universe, rama_ref_array, client_Ramachandran
    ):
        rama = Ramachandran(universe.select_atoms("protein")).run(
            start=5, stop=6, **client_Ramachandran
        )

        assert_allclose(
            rama.results.angles[0],
            rama_ref_array[5],
            rtol=0,
            atol=1.5e-5,
            err_msg="error: dihedral angles should " "match test values",
        )

    def test_ramachandran_residue_selections(
        self, universe, client_Ramachandran
    ):
        rama = Ramachandran(universe.select_atoms("resname GLY")).run(
            **client_Ramachandran
        )
        test_rama = np.load(GLYRamaArray)

        assert_allclose(
            rama.results.angles,
            test_rama,
            rtol=0,
            atol=1.5e-5,
            err_msg="error: dihedral angles should " "match test values",
        )

    def test_outside_protein_length(self, universe, client_Ramachandran):
        with pytest.raises(ValueError):
            rama = Ramachandran(
                universe.select_atoms("resid 220"), check_protein=True
            ).run(**client_Ramachandran)

    def test_outside_protein_unchecked(self, universe, client_Ramachandran):
        rama = Ramachandran(
            universe.select_atoms("resid 220"), check_protein=False
        ).run(**client_Ramachandran)

    def test_protein_ends(self, universe):
        with pytest.warns(UserWarning) as record:
            rama = Ramachandran(
                universe.select_atoms("protein"), check_protein=True
            ).run()
        assert len(record) == 1

    def test_None_removal(self):
        with pytest.warns(UserWarning):
            u = mda.Universe(PDB_rama)
            rama = Ramachandran(u.select_atoms("protein").residues[1:-1])

    def test_plot(self, universe):
        ax = (
            Ramachandran(universe.select_atoms("resid 5-10"))
            .run()
            .plot(ref=True)
        )
        assert isinstance(
            ax, matplotlib.axes.Axes
        ), "Ramachandran.plot() did not return and Axes instance"

    def test_ramachandran_attr_warning(self, universe):
        rama = Ramachandran(universe.select_atoms("protein")).run(stop=2)

        wmsg = "The `angle` attribute was deprecated in MDAnalysis 2.0.0"
        with pytest.warns(DeprecationWarning, match=wmsg):
            assert_equal(rama.angles, rama.results.angles)


class TestJanin(object):

    @pytest.fixture()
    def universe(self):
        return mda.Universe(GRO, XTC)

    @pytest.fixture()
    def universe_tpr(self):
        return mda.Universe(TPR, XTC)

    @pytest.fixture()
    def janin_ref_array(self):
        return np.load(JaninArray)

    def test_janin(self, universe, janin_ref_array, client_Janin):
        self._test_janin(universe, janin_ref_array, client_Janin)

    def test_janin_tpr(self, universe_tpr, janin_ref_array, client_Janin):
        """Test that CYSH are filtered (#2898)"""
        self._test_janin(universe_tpr, janin_ref_array, client_Janin)

    def _test_janin(self, u, ref_array, client_Janin):
        janin = Janin(u.select_atoms("protein")).run(**client_Janin)

        # Test precision lowered to account for platform differences with osx
        assert_allclose(
            janin.results.angles,
            ref_array,
            rtol=0,
            atol=1.5e-3,
            err_msg="error: dihedral angles should " "match test values",
        )

    def test_janin_single_frame(self, universe, janin_ref_array):
        janin = Janin(universe.select_atoms("protein")).run(start=5, stop=6)

        assert_allclose(
            janin.results.angles[0],
            janin_ref_array[5],
            rtol=0,
            atol=1.5e-3,
            err_msg="error: dihedral angles should " "match test values",
        )

    def test_janin_residue_selections(self, universe, client_Janin):
        janin = Janin(universe.select_atoms("resname LYS")).run(**client_Janin)
        test_janin = np.load(LYSJaninArray)

        assert_allclose(
            janin.results.angles,
            test_janin,
            rtol=0,
            atol=1.5e-3,
            err_msg="error: dihedral angles should " "match test values",
        )

    def test_outside_protein_length(self, universe):
        with pytest.raises(ValueError):
            janin = Janin(universe.select_atoms("resid 220")).run()

    def test_remove_residues(self, universe):
        with pytest.warns(UserWarning):
            janin = Janin(universe.select_atoms("protein")).run()

    def test_atom_selection(self):
        with pytest.raises(ValueError):
            u = mda.Universe(PDB_janin)
            janin = Janin(
                u.select_atoms(
                    "protein and not resname ALA CYS GLY " "PRO SER THR VAL"
                )
            )

    def test_plot(self, universe):
        ax = Janin(universe.select_atoms("resid 5-10")).run().plot(ref=True)
        assert isinstance(
            ax, matplotlib.axes.Axes
        ), "Ramachandran.plot() did not return and Axes instance"

    def test_janin_attr_warning(self, universe):
        janin = Janin(universe.select_atoms("protein")).run(stop=2)

        wmsg = "The `angle` attribute was deprecated in MDAnalysis 2.0.0"
        with pytest.warns(DeprecationWarning, match=wmsg):
            assert_equal(janin.angles, janin.results.angles)


# tests for parallelization


@pytest.mark.parametrize(
    "classname,is_parallelizable",
    [
        (MDAnalysis.analysis.dihedrals.Dihedral, True),
        (MDAnalysis.analysis.dihedrals.Ramachandran, True),
        (MDAnalysis.analysis.dihedrals.Janin, True),
    ],
)
def test_class_is_parallelizable(classname, is_parallelizable):
    assert classname._analysis_algorithm_is_parallelizable == is_parallelizable


@pytest.mark.parametrize(
    "classname,backends",
    [
        (
            MDAnalysis.analysis.dihedrals.Dihedral,
            (
                "serial",
                "multiprocessing",
                "dask",
            ),
        ),
        (
            MDAnalysis.analysis.dihedrals.Ramachandran,
            (
                "serial",
                "multiprocessing",
                "dask",
            ),
        ),
        (
            MDAnalysis.analysis.dihedrals.Janin,
            (
                "serial",
                "multiprocessing",
                "dask",
            ),
        ),
    ],
)
def test_supported_backends(classname, backends):
    assert classname.get_supported_backends() == backends



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""\
Trajectory translation --- :mod:`MDAnalysis.transformations.translate`
======================================================================

Translate the coordinates of a given trajectory by a given vector.
The vector can either be user defined, using the function :func:`translate`
or defined by centering an AtomGroup in the unit cell using the function
:func:`center_in_box`

.. autoclass:: translate

.. autoclass:: center_in_box


"""
import numpy as np
from functools import partial

from .base import TransformationBase


class translate(TransformationBase):
    """
    Translates the coordinates of a given :class:`~MDAnalysis.coordinates.timestep.Timestep`
    instance by a given vector.

    Example
    -------
    .. code-block:: python

        ts = MDAnalysis.transformations.translate([1,2,3])(ts)

    Parameters
    ----------
    vector: array-like
        coordinates of the vector to which the coordinates will be translated

    Returns
    -------
    :class:`~MDAnalysis.coordinates.timestep.Timestep` object


    .. versionchanged:: 2.0.0
       The transformation was changed from a function/closure to a class
       with ``__call__``.
    .. versionchanged:: 2.0.0
       The transformation was changed to inherit from the base class for
       limiting threads and checking if it can be used in parallel analysis.
    """

    def __init__(self, vector, max_threads=None, parallelizable=True):
        super().__init__(
            max_threads=max_threads, parallelizable=parallelizable
        )

        self.vector = vector

        if len(self.vector) > 2:
            self.vector = np.float32(self.vector)
        else:
            raise ValueError("{} vector is too short".format(self.vector))

    def _transform(self, ts):
        ts.positions += self.vector
        return ts


class center_in_box(TransformationBase):
    """
    Translates the coordinates of a given :class:`~MDAnalysis.coordinates.timestep.Timestep`
    instance so that the center of geometry/mass of the given :class:`~MDAnalysis.core.groups.AtomGroup`
    is centered on the unit cell. The unit cell dimensions are taken from the input Timestep object.
    If a point is given, the center of the atomgroup will be translated to this point instead.

    Example
    -------

    .. code-block:: python

        ag = u.residues[1].atoms
        ts = MDAnalysis.transformations.center(ag,center='mass')(ts)

    Parameters
    ----------
    ag: AtomGroup
        atom group to be centered on the unit cell.
    center: str, optional
        used to choose the method of centering on the given atom group. Can be 'geometry'
        or 'mass'
    point: array-like, optional
        overrides the unit cell center - the coordinates of the Timestep are translated so
        that the center of mass/geometry of the given AtomGroup is aligned to this position
        instead. Defined as an array of size 3.
    wrap: bool, optional
        If `True`, all the atoms from the given AtomGroup will be moved to the unit cell
        before calculating the center of mass or geometry. Default is `False`, no changes
        to the atom coordinates are done before calculating the center of the AtomGroup.

    Returns
    -------
    :class:`~MDAnalysis.coordinates.timestep.Timestep` object


    .. versionchanged:: 2.0.0
        The transformation was changed from a function/closure to a class
        with ``__call__``.
    .. versionchanged:: 2.0.0
       The transformation was changed to inherit from the base class for
       limiting threads and checking if it can be used in parallel analysis.
    """

    def __init__(
        self,
        ag,
        center="geometry",
        point=None,
        wrap=False,
        max_threads=None,
        parallelizable=True,
    ):
        super().__init__(
            max_threads=max_threads, parallelizable=parallelizable
        )

        self.ag = ag
        self.center = center
        self.point = point
        self.wrap = wrap

        pbc_arg = self.wrap
        if self.point:
            self.point = np.asarray(self.point, np.float32)
            if self.point.shape != (3,) and self.point.shape != (1, 3):
                raise ValueError("{} is not a valid point".format(self.point))
        try:
            if self.center == "geometry":
                self.center_method = partial(
                    self.ag.center_of_geometry, wrap=pbc_arg
                )
            elif self.center == "mass":
                self.center_method = partial(
                    self.ag.center_of_mass, wrap=pbc_arg
                )
            else:
                raise ValueError(f"{self.center} is valid for center")
        except AttributeError:
            if self.center == "mass":
                errmsg = f"{self.ag} is not an AtomGroup object with masses"
                raise AttributeError(errmsg) from None
            else:
                raise ValueError(
                    f"{self.ag} is not an AtomGroup object"
                ) from None

    def _transform(self, ts):
        if self.point is None:
            if ts.dimensions is None:
                raise ValueError("Box is None")
            boxcenter = np.sum(ts.triclinic_dimensions, axis=0) / 2
        else:
            boxcenter = self.point

        ag_center = self.center_method()

        vector = boxcenter - ag_center
        ts.positions += vector

        return ts



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""
Polymer analysis --- :mod:`MDAnalysis.analysis.polymer`
=======================================================


:Author: Richard J. Gowers
:Year: 2015, 2018
:Copyright: Lesser GNU Public License v2.1+

This module contains various commonly used tools in analysing polymers.
"""
import numpy as np
import scipy.optimize
import warnings
import logging

from .. import NoDataError
from ..core.groups import requires, AtomGroup
from ..lib.distances import calc_bonds
from .base import AnalysisBase, ResultsGroup

logger = logging.getLogger(__name__)


@requires("bonds")
def sort_backbone(backbone):
    """Rearrange a linear AtomGroup into backbone order

    Requires that the backbone has bond information,
    and that only backbone atoms are provided (ie no side
    chains or hydrogens).

    Parameters
    ----------
    backbone : AtomGroup
      the backbone atoms, not necessarily in order

    Returns
    -------
    sorted_backbone : AtomGroup
      backbone in order, so `sorted_backbone[i]` is bonded to
      `sorted_backbone[i - 1]` and `sorted_backbone[i + 1]`


    .. versionadded:: 0.20.0
    """
    degrees = [len(atom.bonded_atoms & backbone) for atom in backbone]
    deg1_atoms = [atom for atom, d in zip(backbone, degrees) if d == 1]
    wrong_atoms = [
        atom for atom, d in zip(backbone, degrees) if d not in (1, 2)
    ]

    if len(wrong_atoms) > 0:
        raise ValueError(
            "Backbone contains atoms with connectivity degree not equal to 1 or 2. "
            "This suggests branches or isolated atoms. Problematic atoms: {}."
            "".format(", ".join(str(a) for a in wrong_atoms))
        )

    if len(deg1_atoms) != 2:
        raise ValueError(
            "Backbone connectivity invalid: "
            "expected exactly 2 atoms with connectivity degree 1 (caps). "
            "Cyclical structures are not supported. "
            "Atoms with connectivity degree 1 found: {}."
            "".format(", ".join(str(a) for a in deg1_atoms))
        )

    # arbitrarily choose one of the capping atoms to be the startpoint
    sorted_backbone = AtomGroup([deg1_atoms[0]])

    for _ in range(len(backbone) - 1):
        # current end of the chain
        end_atom = sorted_backbone[-1]

        # look at all bonded atoms which are also part of the backbone
        # and subtract any that have already been added
        next_atom = (end_atom.bonded_atoms & backbone) - sorted_backbone

        # append this to the sorted backbone
        sorted_backbone += next_atom

    return sorted_backbone


class PersistenceLength(AnalysisBase):
    r"""Calculate the persistence length for polymer chains

    The persistence length is the length at which two points on the polymer
    chain become decorrelated.  This is determined by first measuring the
    autocorrelation (:math:`C(n)`) of two bond vectors
    (:math:`\mathbf{a}_i, \mathbf{a}_{i + n}`) separated by :math:`n` bonds

    .. math::

       C(n) = \langle \cos\theta_{i, i+n} \rangle =
               \langle \mathbf{a_i} \cdot \mathbf{a_{i+n}} \rangle

    where :math:`a_i` and :math:`a_{i+n}` are unit vectors
    along the bonds.

    An exponential decay is then fitted to this, which yields the
    persistence length

    .. math::

       C(n) \approx \exp\left( - \frac{n \bar{l_B}}{l_P} \right)

    where :math:`\bar{l_B}` is the average bond length, and :math:`l_P` is
    the persistence length which is fitted

    Parameters
    ----------
    atomgroups : iterable
       List of AtomGroups. Each should represent a single
       polymer chain, ordered in the correct order.
    verbose : bool, optional
       Show detailed progress of the calculation if set to ``True``.

    Attributes
    ----------
    results.bond_autocorrelation : numpy.ndarray
       the measured bond autocorrelation
    results.lb : float
       the average bond length

       .. versionadded:: 2.0.0

    lb : float

       Alias to the :attr:`results.lb`.

       .. deprecated:: 2.0.0
            Will be removed in MDAnalysis 3.0.0. Please use
            :attr:`results.lb` instead.

    results.x : numpy.ndarray
        length of the decorrelation predicted by *lp*

        .. versionadded:: 2.0.0

    results.lp : float
       calculated persistence length

       .. versionadded:: 2.0.0

    lp : float

       Alias to the :attr:`results.lp`.

       .. deprecated:: 2.0.0
            Will be removed in MDAnalysis 3.0.0. Please use
            :attr:`results.lp` instead.

    results.fit : numpy.ndarray
       the modelled backbone decorrelation predicted by *lp*

       .. versionadded:: 2.0.0

    fit : float

       Alias to the :attr:`results.fit`.

       .. deprecated:: 2.0.0
            Will be removed in MDAnalysis 3.0.0. Please use
            :attr:`results.fit` instead.

    See Also
    --------
    :func:`sort_backbone`
       for producing the sorted AtomGroup required for input.

    Example
    -------
    .. code-block:: python

        from MDAnalysis.tests.datafiles import TRZ_psf, TRZ
        import MDAnalysis as mda
        from MDAnalysis.analysis import polymer
        u = mda.Universe(TRZ_psf, TRZ)

        # this system is a pure polymer melt of polyamide,
        # so we can select the chains by using the .fragments attribute
        chains = u.atoms.fragments

        # select only the backbone atoms for each chain
        backbones = [chain.select_atoms('not name O* H*') for chain in chains]

        # sort the chains, removing any non-backbone atoms
        sorted_backbones = [polymer.sort_backbone(bb) for bb in backbones]
        persistence_length = polymer.PersistenceLength(sorted_backbones)

        # Run the analysis, this will average over all polymer chains
        # and all timesteps in trajectory
        persistence_length = persistence_length.run()
        print(f'The persistence length is: {persistence_length.results.lp}')

        # always check the visualisation of this:
        persistence_length.plot()


    .. versionadded:: 0.13.0
    .. versionchanged:: 0.20.0
       The run method now automatically performs the exponential fit
    .. versionchanged:: 1.0.0
       Deprecated :meth:`PersistenceLength.perform_fit` has now been removed.
    .. versionchanged:: 2.0.0
       Former ``results`` are now stored as ``results.bond_autocorrelation``.
       :attr:`lb`, :attr:`lp`, :attr:`fit` are now stored in a
       :class:`MDAnalysis.analysis.base.Results` instance.
    .. versionchanged:: 2.10.0
       introduced :meth:`get_supported_backends` allowing for parallel
       execution on ``multiprocessing`` and ``dask`` backends.
    """

    _analysis_algorithm_is_parallelizable = True

    @classmethod
    def get_supported_backends(cls):
        return ("serial", "multiprocessing", "dask")

    def __init__(self, atomgroups, **kwargs):
        super(PersistenceLength, self).__init__(
            atomgroups[0].universe.trajectory, **kwargs
        )
        self._atomgroups = atomgroups

        # Check that all chains are the same length
        lens = [len(ag) for ag in atomgroups]
        chainlength = len(atomgroups[0])
        if not all(l == chainlength for l in lens):
            raise ValueError("Not all AtomGroups were the same size")
        self.chainlength = chainlength

    def _prepare(self):
        self.results.raw_bond_autocorr = np.zeros(
            self.chainlength - 1, dtype=np.float32
        )

    def _single_frame(self):
        # could optimise this by writing a "self dot array"
        # we're only using the upper triangle of np.inner
        # function would accept a bunch of coordinates and spit out the
        # decorrel for that

        for chain in self._atomgroups:
            # Vector from each atom to next
            vecs = chain.positions[1:] - chain.positions[:-1]
            # Normalized to unit vectors
            vecs /= np.sqrt((vecs * vecs).sum(axis=1))[:, None]

            inner_pr = np.inner(vecs, vecs)
            for i in range(self.chainlength - 1):
                self.results.raw_bond_autocorr[
                    : (self.chainlength - 1) - i
                ] += inner_pr[i, i:]

    def _get_aggregator(self):
        return ResultsGroup(
            lookup={
                "raw_bond_autocorr": ResultsGroup.ndarray_sum,
            }
        )

    @property
    def lb(self):
        wmsg = (
            "The `lb` attribute was deprecated in "
            "MDAnalysis 2.0.0 and will be removed in MDAnalysis 3.0.0. "
            "Please use `results.variance` instead."
        )
        warnings.warn(wmsg, DeprecationWarning)
        return self.results.lb

    @property
    def lp(self):
        wmsg = (
            "The `lp` attribute was deprecated in "
            "MDAnalysis 2.0.0 and will be removed in MDAnalysis 3.0.0. "
            "Please use `results.variance` instead."
        )
        warnings.warn(wmsg, DeprecationWarning)
        return self.results.lp

    @property
    def fit(self):
        wmsg = (
            "The `fit` attribute was deprecated in "
            "MDAnalysis 2.0.0 and will be removed in MDAnalysis 3.0.0. "
            "Please use `results.variance` instead."
        )
        warnings.warn(wmsg, DeprecationWarning)
        return self.results.fit

    def _conclude(self):
        norm = np.linspace(self.chainlength - 1, 1, self.chainlength - 1)
        norm *= len(self._atomgroups) * self._trajectory.n_frames
        self.results.bond_autocorrelation = (
            self.results.raw_bond_autocorr / norm
        )
        self._calc_bond_length()
        self._perform_fit()

    def _calc_bond_length(self):
        """calculate average bond length"""
        bs = []
        for ag in self._atomgroups:
            pos = ag.positions
            b = calc_bonds(pos[:-1], pos[1:]).mean()
            bs.append(b)
        self.results.lb = np.mean(bs)

    def _perform_fit(self):
        """Fit the results to an exponential decay"""
        try:
            self.results.bond_autocorrelation
        except AttributeError:
            raise NoDataError("Use the run method first") from None
        self.results.x = self.results.lb * np.arange(
            len(self.results.bond_autocorrelation)
        )

        self.results.lp = fit_exponential_decay(
            self.results.x, self.results.bond_autocorrelation
        )

        self.results.fit = np.exp(-self.results.x / self.results.lp)

    def plot(self, ax=None):
        """Visualize the results and fit

        Parameters
        ----------
        ax : matplotlib.Axes, optional
          if provided, the graph is plotted on this axis

        Returns
        -------
        ax : the axis that the graph was plotted on
        """
        import matplotlib.pyplot as plt

        if ax is None:
            _, ax = plt.subplots()
        ax.plot(
            self.results.x,
            self.results.bond_autocorrelation,
            "ro",
            label="Result",
        )
        ax.plot(self.results.x, self.results.fit, label="Fit")
        ax.set_xlabel(r"x")
        ax.set_ylabel(r"$C(x)$")
        ax.set_xlim(0.0, 40 * self.results.lb)

        ax.legend(loc="best")

        return ax


def fit_exponential_decay(x, y):
    r"""Fit a function to an exponential decay

    .. math::  y = \exp\left(- \frac{x}{a}\right)

    Parameters
    ----------
    x, y : array_like
      The two arrays of data

    Returns
    -------
    a : float
      The coefficient *a* for this decay

    Notes
    -----
    This function assumes that data starts at 1.0 and decays to 0.0

    """

    def expfunc(x, a):
        return np.exp(-x / a)

    a = scipy.optimize.curve_fit(expfunc, x, y)[0][0]

    return a



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#


"""
CHARMM selections
=================

Write :class:`MDAnalysis.core.groups.AtomGroup` selection to a `str` file
that defines a `CHARMM selection`_. To be used in CHARMM_ like this::

  stream macro.str

The selection is named *mdanalysis001*.

.. autoclass:: SelectionWriter
   :inherited-members:

.. _CHARMM: http://www.charmm.org
.. _CHARMM selection: http://www.charmm.org/documentation/c34b1/select.html
"""
from . import base


class SelectionWriter(base.SelectionWriterBase):
    format = ["CHARMM", "str"]
    ext = "str"
    continuation = "-"
    commentfmt = "! %s"
    default_numterms = (
        4  # be conservative because CHARMM only reads 72 columns
    )

    def _translate(self, atoms, **kwargs):
        # CHARMM index is 1-based
        def _index(atom):
            return "BYNUM {0:d}".format((atom.index + 1))

        return base.join(atoms, " .or.", _index)

    def _write_head(self, out, **kwargs):
        out.write(self.comment("MDAnalysis CHARMM selection"))
        out.write(
            "DEFINE {name!s} SELECT ".format(**kwargs)
            + self.continuation
            + "\n"
        )

    def _write_tail(self, out, **kwargs):
        out.write("END")



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
"""
Cython wrapper for the C implementation of the Stochastic Proximity Embedding
dimensionality reduction algorithm.

:Author: Matteo Tiberti, Wouter Boomsma
"""


import logging
import numpy
cimport numpy

numpy.import_array()
cimport cython

cdef extern from "spe.h":
    double CStochasticProximityEmbedding(double*, double*, double, int, int, double, double, int, int, int)


@cython.embedsignature(True)
def StochasticProximityEmbedding(s, double rco, int dim, double maxlam, double minlam, int ncycle, int nstep, int stressfreq):
    """
    Stochastic proximity embedding dimensionality reduction algorithm. The
    algorithm implemented here is described in this paper:

    Dmitrii N. Rassokhin, Dimitris K. Agrafiotis
    A modified update rule for stochastic proximity embedding
    Journal of Molecular Graphics and Modelling 22 (2003) 133–140

    This class is a Cython wrapper for a C implementation (see spe.c)

    Parameters
    ----------
    s : encore.utils.TriangularMatrix object
        Triangular matrix containing the distance values for each pair of
        elements in the original space.
    rco : float
        neighborhood distance cut-off
    dim : int
        number of dimensions for the embedded space
    minlam  : float
        final learning parameter
    maxlam  : float
        starting learning parameter
    ncycle : int
        number of cycles. Each cycle is composed of nstep steps. At the end
        of each cycle, the lerning parameter lambda is updated.
    nstep : int
        number of coordinate update steps for each cycle


    Returns
    -------
    space : (float, numpy.array)
        float is the final stress obtained; the array are the coordinates of
        the elements in the embedded space
    stressfreq : int
        calculate and report stress value every stressfreq cycle


    """

    cdef int nelem = s.size
    cdef double finalstress = 0.0

    logging.info("Starting Stochastic Proximity Embedding")

    cdef numpy.ndarray[numpy.float64_t,  ndim=1] matndarray = numpy.ascontiguousarray(s._elements, dtype=numpy.float64)
    cdef numpy.ndarray[numpy.float64_t,  ndim=1] d_coords   = numpy.zeros((nelem*dim),dtype=numpy.float64)

    finalstress = CStochasticProximityEmbedding(<double*>matndarray.data, <double*>d_coords.data, rco, nelem, dim, maxlam, minlam, ncycle, nstep, stressfreq)

    logging.info("Stochastic Proximity Embedding finished. Residual stress: %.3f" % finalstress)

    return (finalstress, d_coords.reshape((-1,dim)).T)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""
:mod:`MDAnalysis` --- analysis of molecular simulations in python
=================================================================

MDAnalysis (https://www.mdanalysis.org) is a python toolkit to analyze
molecular dynamics trajectories generated by CHARMM, NAMD, Amber,
Gromacs, or LAMMPS.

It allows one to read molecular dynamics trajectories and access the
atomic coordinates through numpy arrays. This provides a flexible and
relatively fast framework for complex analysis tasks. In addition,
CHARMM-style atom selection commands are implemented. Trajectories can
also be manipulated (for instance, fit to a reference structure) and
written out. Time-critical code is written in C for speed.

Help is also available through GitHub Discussions at
https://github.com/MDAnalysis/mdanalysis/discussions

Please report bugs and feature requests through the issue tracker at
https://github.com/MDAnalysis/mdanalysis/issues

Citation
--------

When using MDAnalysis in published work, please cite

    R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
    D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein. 
    MDAnalysis: A Python package for the rapid analysis of molecular dynamics 
    simulations. In S. Benthall and S. Rostrup, editors, Proceedings of the 15th 
    Python in Science Conference, pages 98-105, Austin, TX, 2016. SciPy, 
    doi:10.25080/majora-629e541a-00e

    N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and
    O. Beckstein. MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics
    Simulations. J. Comput. Chem. 32 (2011), 2319--2327, doi:`10.1002/jcc.21787`_
    https://www.mdanalysis.org

For citations of included algorithms and sub-modules please see the references_.

.. _`10.1002/jcc.21787`: http://dx.doi.org/10.1002/jcc.21787
.. _references: https://docs.mdanalysis.org/documentation_pages/references.html


Getting started
---------------

Import the package::

  >>> import MDAnalysis

(note that not everything in MDAnalysis is imported right away; for
additional functionality you might have to import sub-modules
separately, e.g. for RMS fitting ``import MDAnalysis.analysis.align``.)

Build a "universe" from a topology (PSF, PDB) and a trajectory (DCD, XTC/TRR);
here we are assuming that PSF, DCD, etc contain file names. If you don't have
trajectories at hand you can play with the ones that come with MDAnalysis for
testing (see below under `Examples`_)::

  >>> u = MDAnalysis.Universe(PSF, DCD)

Select the C-alpha atoms and store them as a group of atoms::

  >>> ca = u.select_atoms('name CA')
  >>> len(ca)
  214

Calculate the centre of mass of the CA and of all atoms::

  >>> ca.center_of_mass()
  array([ 0.06873595, -0.04605918, -0.24643682])
  >>> u.atoms.center_of_mass()
  array([-0.01094035,  0.05727601, -0.12885778])

Calculate the CA end-to-end distance (in angstroem)::
  >>> import numpy as np
  >>> coord = ca.positions
  >>> v = coord[-1] - coord[0]   # last Ca minus first one
  >>> np.sqrt(np.dot(v, v,))
  10.938133

Define a function eedist():
  >>> def eedist(atoms):
  ...     coord = atoms.positions
  ...     v = coord[-1] - coord[0]
  ...     return sqrt(dot(v, v,))
  ...
  >>> eedist(ca)
  10.938133

and analyze all timesteps *ts* of the trajectory::
  >>> for ts in u.trajectory:
  ...      print eedist(ca)
  10.9381
  10.8459
  10.4141
   9.72062
  ....

See Also
--------
:class:`MDAnalysis.core.universe.Universe` for details


Examples
--------

MDAnalysis comes with a number of real trajectories for testing. You
can also use them to explore the functionality and ensure that
everything is working properly::

  from MDAnalysis import *
  from MDAnalysis.tests.datafiles import PSF,DCD, PDB,XTC
  u_dims_adk = Universe(PSF,DCD)
  u_eq_adk = Universe(PDB, XTC)

The PSF and DCD file are a closed-form-to-open-form transition of
Adenylate Kinase (from [Beckstein2009]_) and the PDB+XTC file are ten
frames from a Gromacs simulation of AdK solvated in TIP4P water with
the OPLS/AA force field.

.. [Beckstein2009] O. Beckstein, E.J. Denning, J.R. Perilla and T.B. Woolf,
   Zipping and Unzipping of Adenylate Kinase: Atomistic Insights into the
   Ensemble of Open <--> Closed Transitions. J Mol Biol 394 (2009), 160--176,
   doi:10.1016/j.jmb.2009.09.009

"""

__all__ = ["Universe", "Writer", "AtomGroup", "ResidueGroup", "SegmentGroup"]

import logging
import warnings
from typing import Dict


logger = logging.getLogger("MDAnalysis.__init__")

from .version import __version__

try:
    from .authors import __authors__
except ImportError:
    logger.info("Could not find authors.py, __authors__ will be empty.")
    __authors__ = []

# Registry of Readers, Parsers and Writers known to MDAnalysis
# Metaclass magic fills these as classes are declared.
_READERS: Dict = {}
_READER_HINTS: Dict = {}
_SINGLEFRAME_WRITERS: Dict = {}
_MULTIFRAME_WRITERS: Dict = {}
_PARSERS: Dict = {}
_PARSER_HINTS: Dict = {}
_SELECTION_WRITERS: Dict = {}
_CONVERTERS: Dict = {}
# Registry of TopologyAttributes
_TOPOLOGY_ATTRS: Dict = {}  # {attrname: cls}
_TOPOLOGY_TRANSPLANTS: Dict = (
    {}
)  # {name: [attrname, method, transplant class]}
_TOPOLOGY_ATTRNAMES: Dict = {}  # {lower case name w/o _ : name}
_GUESSERS: Dict = {}

# custom exceptions and warnings
from .exceptions import (
    SelectionError,
    NoDataError,
    ApplicationError,
    SelectionWarning,
    MissingDataWarning,
    ConversionWarning,
    FileFormatWarning,
    StreamWarning,
)

from .lib import log
from .lib.log import start_logging, stop_logging

logging.getLogger("MDAnalysis").addHandler(log.NullHandler())
del logging

# only MDAnalysis DeprecationWarnings are loud by default
warnings.filterwarnings(
    action="once", category=DeprecationWarning, module="MDAnalysis"
)


from . import units

# Bring some often used objects into the current namespace
from .core.universe import Universe, Merge
from .core.groups import AtomGroup, ResidueGroup, SegmentGroup
from .coordinates.core import writer as Writer

# After Universe import
from . import converters

from .due import due, Doi, BibTeX

due.cite(
    Doi("10.25080/majora-629e541a-00e"),
    description="Molecular simulation analysis library",
    path="MDAnalysis",
    cite_module=True,
)
due.cite(
    Doi("10.1002/jcc.21787"),
    description="Molecular simulation analysis library",
    path="MDAnalysis",
    cite_module=True,
)

del Doi, BibTeX



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""\
No Jump Trajectory Unwrapping --- :mod:`MDAnalysis.transformations.nojump`
=======================================================================================

Unwraps the trajectory such that atoms never move more than half a periodic box length.
The consequence of this is that particles will diffuse across periodic boundaries when
needed. This unwrapping method is suitable as a preprocessing step to calculate
molecular diffusion, or more commonly to keep multi-domain proteins whole during trajectory
analysis.
The algorithm used is based on :footcite:p:`Kulke2022`.

.. autoclass:: NoJump

"""
import numpy as np
import warnings

from .base import TransformationBase
from ..due import due, Doi
from ..exceptions import NoDataError


class NoJump(TransformationBase):
    """
    Returns transformed coordinates for the given timestep so that an atom
    does not move more than half the periodic box size between two timesteps, and will move
    across periodic boundary edges. The algorithm used is based on :footcite:p:`Kulke2022`,
    equation B6 for non-orthogonal systems, so it is general to most applications where
    molecule trajectories should not "jump" from one side of a periodic box to another.

    Note that this transformation depends on a periodic box dimension being set for every
    frame in the trajectory, and that this box dimension can be transformed to an orthonormal
    unit cell. If not, an error is emitted. Since it is typical to transform all frames
    sequentially when unwrapping trajectories, warnings are emitted if non-sequential timesteps
    are transformed using NoJump.

    Example
    -------

    Suppose you had a molecular trajectory with a protein dimer that moved across a periodic
    boundary. This will normally appear to make the dimer split apart. This transformation
    uses the molecular motions between frames in a wrapped trajectory to create an unwrapped
    trajectory where molecules never move more than half a periodic box dimension between subsequent
    frames. Thus, the normal use case is to apply the transformation to every frame of a trajectory,
    and to do so sequentially.

    .. code-block:: python

        transformation = NoJump()
        u.trajectory.add_transformations(transformation)
        for ts in u.trajectory:
            print(ts.positions)

    In this case, ``ts.positions`` will return the NoJump unwrapped trajectory, which would keep
    protein dimers together if they are together in the inital timestep. The unwrapped trajectory may
    also be useful to compute diffusion coefficients via the Einstein relation.
    To reverse the process, wrap the coordinates again.

    Returns
    -------
    MDAnalysis.coordinates.timestep.Timestep

    References
    ----------
    .. footbibliography::

    """

    @due.dcite(
        Doi("10.1021/acs.jctc.2c00327"),
        description="Works through the orthogonal case for unwrapping, "
        "and proposes the non-orthogonal approach.",
        path=__name__,
    )
    def __init__(
        self,
        check_continuity=True,
        max_threads=None,
    ):
        # NoJump transforms are inherently unparallelizable, since
        # it depends on the previous frame's unwrapped coordinates
        super().__init__(max_threads=max_threads, parallelizable=False)
        self.prev = None
        self.old_frame = 0
        self.older_frame = "A"
        self.check_c = check_continuity

    def _transform(self, ts):
        L = ts.triclinic_dimensions
        if L is None:
            msg = f"Periodic box dimensions not provided at step {ts.frame}"
            raise NoDataError(msg)
        try:
            Linverse = np.linalg.inv(L)
        except np.linalg.LinAlgError:
            msg = f"Periodic box dimensions are not invertible at step {ts.frame}"
            raise NoDataError(msg)
        if ts.frame == 0:
            # We don't need to apply the transformation here. However, we need to
            # ensure we have the 0th frame coordinates in reduced form. We also need to
            # set an an appropriate value for self.older frame. This is so that on the
            # following frame we don't return early when we check
            # `self.older_frame != "A"`. If we return early, then the transformation is
            # not applied, and any jumps across boundaries that occur at that frame will
            # not be accounted for.
            self.prev = ts.positions @ Linverse
            self.old_frame = 0
            self.older_frame = -1
            return ts
        if (
            self.check_c
            and self.older_frame != "A"
            and (self.old_frame - self.older_frame)
            != (ts.frame - self.old_frame)
        ):
            warnings.warn(
                "NoJump detected that the interval between frames is unequal."
                "We are resetting the reference frame to the current timestep.",
                UserWarning,
            )
            self.prev = ts.positions @ Linverse
            self.old_frame = ts.frame
            self.older_frame = "A"
            return ts
        if self.check_c and np.abs(self.old_frame - ts.frame) != 1:
            warnings.warn(
                "NoJump transform is only accurate when positions"
                "do not move by more than half a box length."
                "Currently jumping between frames with a step of more than 1."
                "This might be fine, but depending on the trajectory stride,"
                "this might be inaccurate.",
                UserWarning,
            )
        # Convert into reduced coordinate space
        fcurrent = ts.positions @ Linverse
        fprev = (
            self.prev
        )  # Previous unwrapped coordinates in reduced box coordinates.
        # Calculate the new positions in reduced coordinate space (Equation B6 from
        # 10.1021/acs.jctc.2c00327). As it turns out, the displacement term can
        # be moved inside the round function in this coordinate space, as the
        # difference between wrapped and unwrapped coordinates is an integer.
        newpositions = fcurrent - np.round(fcurrent - fprev)
        # Convert back into real space
        ts.positions = newpositions @ L
        # Set things we need to save for the next frame.
        self.prev = (
            newpositions  # Note that this is in reduced coordinate space.
        )
        self.older_frame = self.old_frame
        self.old_frame = ts.frame

        return ts



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import re

import numpy as np
import pytest
from numpy.testing import assert_equal, assert_almost_equal, assert_allclose

import MDAnalysis as mda
from MDAnalysis.analysis import helix_analysis as hel
from MDAnalysisTests.datafiles import (
    GRO,
    XTC,
    PSF,
    DCD,
    PDB_small,
    HELANAL_BENDING_MATRIX,
    HELANAL_BENDING_MATRIX_SUBSET,
    XYZ,
)

# reference data from old helix analysis of a single PDB file:
#   data = MDAnalysis.analysis.helanal.helanal_main(PDB_small,
#                    select="name CA and resnum 161-187")
# keys are renamed and local screw angles now use a different
# algorithm
HELANAL_SINGLE_DATA = {
    "global_tilts": np.rad2deg(1.3309656332019535),
    "local_heights summary": np.array(
        [1.5286051, 0.19648294, 0.11384312], dtype=np.float32
    ),
    "local_bends": np.array(
        [
            3.44526005,
            4.85425806,
            4.69548464,
            2.39473653,
            3.56172442,
            3.97128344,
            3.41563916,
            1.86140978,
            5.22997046,
            5.41381264,
            27.49601364,
            39.69839478,
            35.05921936,
            21.78928566,
            9.8632431,
            8.80066967,
            5.5344553,
            6.14356709,
            10.15450764,
            11.07686138,
            9.23541832,
        ],
        dtype=np.float32,
    ),
    "local_nres_per_turn summary": np.array([3.64864163, 0.152694, 0.1131402]),
    "local_twists summary": np.array(
        [98.83011627, 4.08171701, 3.07253003], dtype=np.float32
    ),
    "local_twists": np.array(
        [
            97.23709869,
            99.09676361,
            97.25350952,
            101.76019287,
            100.42689514,
            97.08784485,
            97.07430267,
            98.33553314,
            97.86578369,
            95.45792389,
            97.10089111,
            95.26415253,
            87.93136597,
            108.38458252,
            95.27167511,
            104.01931763,
            100.7199707,
            101.48034668,
            99.64170074,
            94.78946686,
            102.50147247,
            97.25154877,
            104.54204559,
            101.42829895,
        ],
        dtype=np.float32,
    ),
}


def read_bending_matrix(fn):
    """Read helanal_bending_matrix.dat into dict of numpy arrays.

    This is a quick and dirty parser with no error checking.

    Format::

      Mean
         0.0    5.7   10.9 ....
         .

      SD
        ....
        .

      ABDEV
       ....
       .

    Returns
    -------
       data : dict
           dictionary with keys 'Mean', 'SD', 'ABDEV' and NxN matrices.
    """
    data = {}
    current = None
    with open(fn) as bendingmatrix:
        for line in bendingmatrix:
            line = line.strip()
            if not line:
                continue
            if re.match(r"\D+", line):
                label = line.split()[0]
                current = data[label] = []
            else:
                current.append([float(x) for x in line.split()])
    for k, v in data.items():
        data[k] = np.array(v)
    return data


def test_local_screw_angles_plane_circle():
    """
    Test the trivial case of a circle in the xy-plane.
    The global axis is the x-axis and ref axis is the y-axis,
    so angles should be calculated to the xy-plane.
    """
    angdeg = np.arange(0, 360, 12, dtype=np.int32)
    angrad = np.deg2rad(angdeg, dtype=np.float64)
    xyz = np.array(
        [[np.cos(a), np.sin(a), 0] for a in angrad], dtype=np.float64
    )
    xyz[15, 1] = 0  # because np.sin(np.deg2rad(180)) = 1e-16 ?!
    screw = hel.local_screw_angles([1, 0, 0], [0, 1, 0], xyz)
    correct = np.zeros_like(angdeg)
    correct[(angdeg > 180)] = 180
    assert_almost_equal(screw, correct)


def test_local_screw_angles_ortho_circle():
    """
    Test the trivial case of a circle in the xy-plane.
    The global axis is the x-axis and ref axis is the z-axis,
    so angles should be calculated to the xz-plane.
    """
    angdeg = np.arange(0, 360, 12, dtype=np.int32)
    angrad = np.deg2rad(angdeg, dtype=np.float64)
    xyz = np.array(
        [[np.cos(a), np.sin(a), 0] for a in angrad], dtype=np.float64
    )
    xyz[15, 1] = 0  # because np.sin(np.deg2rad(180)) = 1e-16 ?!
    screw = hel.local_screw_angles([1, 0, 0], [0, 0, 1], xyz)
    correct = np.zeros_like(angdeg)
    correct[(angdeg < 180)] = 90
    correct[(angdeg > 180)] = -90
    correct[0] = correct[15] = 0
    assert_almost_equal(screw, correct)


def test_local_screw_angles_around_circle():
    """
    Test an orthogonal circle in the xz-plane.
    The global axis is the y-axis and ref axis is the x-axis,
    so angles should be calculated to the xy-plane.
    """
    # circle in xz-plane
    angdeg = np.arange(0, 360, 12, dtype=int)
    angrad = np.deg2rad(angdeg)
    xyz = np.array([[np.cos(a), 0, np.sin(a)] for a in angrad])
    screw = hel.local_screw_angles([0, 1, 0], [1, 0, 0], xyz)
    angdeg[-14:] = -angdeg[1:15][::-1]
    angdeg[-15] = 180
    assert_almost_equal(screw, angdeg)


def test_local_screw_angles_around_circle_rising():
    """
    Test a circle in the xz-plane, rising on the y-axis.
    The global axis is the y-axis and ref axis is the x-axis,
    so angles should be calculated to the xy-plane.
    The y-axis contribution should be removed so angles should
    be to the circle.
    """
    # circle in xz-plane
    angdeg = np.arange(0, 360, 12, dtype=int)
    angrad = np.deg2rad(angdeg)
    xyz = np.array([[np.cos(a), i, np.sin(a)] for i, a in enumerate(angrad)])
    screw = hel.local_screw_angles([0, 1, 0], [1, 0, 0], xyz)
    angdeg[-14:] = -angdeg[1:15][::-1]
    angdeg[-15] = 180
    assert_almost_equal(screw, angdeg)


def test_local_screw_angles_parallel_axes():
    """
    Test that if global_axis and ref_axis are parallel, it
    picks another one instead. global_axis is the x-axis,
    so the eventual replacement ref_axis should be the z-axis,
    so angles should be calculated to the xz-plane.
    """
    xyz = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
    angles = hel.local_screw_angles([1, 0, 0], [-1, 0, 0], xyz)
    assert_almost_equal(angles, [0, 90, 0])


@pytest.fixture()
def zigzag():
    r"""
    x-axis is from bottom to up in plane of screen
    z-axis is left to right ->
         x    x    x    x    x
        / \  / \  / \  / \  /
      x    x    x    x    x
    """
    n_atoms = 100
    u = mda.Universe.empty(
        100, atom_resindex=np.arange(n_atoms), trajectory=True
    )
    xyz = np.array(
        list(
            zip(
                [1, -1] * (n_atoms // 2),  # x \in {0, 1}
                [0] * n_atoms,  # y == 0
                range(n_atoms),
            )
        )
    )  # z rises continuously
    u.load_new(xyz)
    return u


@pytest.mark.parametrize(
    "ref_axis,screw_angles",
    [
        # input vectors zigzag between [-1, 0, 0] and [1, 0, 0]
        # global axis is z-axis
        ([0, 0, 1], [180, 0]),  # calculated to x-z plane
        ([1, 0, 0], [180, 0]),  # calculated to x-z plane
        ([-1, 0, 0], [0, 180]),  # calculated to x-z plane upside down
        ([0, 1, 0], [90, -90]),  # calculated to y-z plane
        ([0, -1, 0], [-90, 90]),  # calculated to x-z plane upside down
        # calculated to diagonal xy-z plane rotating around
        ([1, 1, 0], [135, -45]),
        ([-1, 1, 0], [45, -135]),
        ([1, -1, 0], [-135, 45]),
        ([-1, -1, 0], [-45, 135]),
        # calculated to diagonal xyz-z plane w/o z contribution
        ([1, 1, 1], [135, -45]),
        ([1, 1, -1], [135, -45]),
        ([1, -1, 1], [-135, 45]),
        ([-1, 1, 1], [45, -135]),
        ([-1, -1, 1], [-45, 135]),
        ([-1, -1, -1], [-45, 135]),
    ],
)
def test_helix_analysis_zigzag(zigzag, ref_axis, screw_angles):
    properties = hel.helix_analysis(zigzag.atoms.positions, ref_axis=ref_axis)
    assert_almost_equal(properties["local_twists"], 180, decimal=4)
    assert_almost_equal(properties["local_nres_per_turn"], 2, decimal=4)
    assert_almost_equal(properties["global_axis"], [0, 0, -1], decimal=4)
    # all 0 vectors
    assert_almost_equal(properties["local_axes"], 0, decimal=4)
    assert_almost_equal(properties["local_bends"], 0, decimal=4)
    assert_almost_equal(properties["all_bends"], 0, decimal=4)
    assert_almost_equal(properties["local_heights"], 0, decimal=4)
    assert_almost_equal(
        properties["local_helix_directions"][0::2],
        [[-1, 0, 0]] * 49,
        decimal=4,
    )
    assert_almost_equal(
        properties["local_helix_directions"][1::2], [[1, 0, 0]] * 49, decimal=4
    )
    origins = zigzag.atoms.positions[1:-1].copy()
    origins[:, 0] = 0
    assert_almost_equal(properties["local_origins"], origins, decimal=4)
    assert_almost_equal(
        properties["local_screw_angles"], screw_angles * 49, decimal=4
    )


def square_oct(n_rep=10):
    """
    square-octagon-square-octagon

    y- and z- coordinates create the square and octogon.
    x-coordinates increase continuously.
    """
    # square-octagon-square-octagon
    sq2 = 0.5**0.5
    square = [(1, 0), (0, 1), (-1, 0), (0, -1)]
    octagon = [
        (1, 0),
        (sq2, sq2),
        (0, 1),
        (-sq2, sq2),
        (-1, 0),
        (-sq2, -sq2),
        (0, -1),
        (sq2, -sq2),
    ]
    shapes = (square + octagon) * n_rep
    xyz = np.array(list(zip(np.arange(len(shapes)), *zip(*shapes))))
    n_atoms = len(xyz)
    u = mda.Universe.empty(n_atoms, trajectory=True)
    u.load_new(xyz)
    return u


def test_helix_analysis_square_oct():
    n_rep = 10
    u = square_oct(n_rep=n_rep)
    n_atoms = len(u.atoms)

    properties = hel.helix_analysis(u.atoms.positions, ref_axis=[0, 0, 1])
    twist_trans = [102.76438, 32.235607]
    twists = ([90] * 2 + twist_trans + [45] * 6 + twist_trans[::-1]) * n_rep
    assert_almost_equal(
        properties["local_twists"], twists[: n_atoms - 3], decimal=4
    )
    res_trans = [3.503159, 11.167775]
    res = ([4] * 2 + res_trans + [8] * 6 + res_trans[::-1]) * n_rep
    assert_almost_equal(
        properties["local_nres_per_turn"], res[: n_atoms - 3], decimal=4
    )
    assert_almost_equal(properties["global_axis"], [-1, 0, 0], decimal=3)
    assert_almost_equal(properties["local_axes"] - [1, 0, 0], 0, decimal=4)
    assert_almost_equal(properties["local_bends"], 0, decimal=4)
    assert_almost_equal(properties["all_bends"], 0, decimal=4)
    assert_almost_equal(properties["local_heights"], 1, decimal=4)

    loc_rot = [
        [0.0, 0.0, 1.0],
        [0.0, -1.0, 0.0],
        [0.0, 0.0, -1.0],
        [0.0, 0.97528684, 0.2209424],  # the transition square->oct
        [0.0, 0.70710677, 0.70710677],  # 0.5 ** 0.5
        [0.0, 0.0, 1.0],
        [0.0, -0.70710677, 0.70710677],
        [0.0, -1.0, 0.0],
        [0.0, -0.70710677, -0.70710677],
        [0.0, 0.0, -1.0],
        [0.0, 0.70710677, -0.70710677],
        [0.0, 0.97528684, -0.2209424],
    ] * n_rep
    assert_almost_equal(
        properties["local_helix_directions"], loc_rot[: n_atoms - 2], decimal=4
    )

    origins = u.atoms.positions.copy()[1:-1]
    origins[:, 1:] = (
        [
            [0.0, 0.0],  # square
            [0.0, 0.0],
            [0.0, -0.33318555],  # square -> octagon
            [-1.7878988, -0.6315732],  # square -> octagon
            [0.0, 0.0],  # octagon
            [0.0, 0.0],
            [0.0, 0.0],
            [0.0, 0.0],
            [0.0, 0.0],
            [0.0, 0.0],
            [-1.3141878, 1.3141878],  # octagon -> square
            [0.34966463, 0.14732757],
        ]
        * n_rep
    )[: len(origins)]
    assert_almost_equal(properties["local_origins"], origins, decimal=4)

    # calculated to the x-y plane
    # all input vectors (loc_rot) are in y-z plane
    screw = [
        0,
        90,
        180,  # square
        -77.236,  # transition
        -45,
        0,
        45,
        90,
        135,
        180,
        -135,  # octagon
        -102.764,
    ] * n_rep

    # not quite 0, comes out as 1.32e-06
    assert_almost_equal(
        properties["local_screw_angles"], screw[:-2], decimal=3
    )


class TestHELANAL(object):

    @pytest.fixture()
    def psf_ca(self):
        u = mda.Universe(PSF, DCD)
        ag = u.select_atoms("name CA")
        return ag

    @pytest.fixture()
    def helanal(self, psf_ca):
        ha = hel.HELANAL(
            psf_ca, select="resnum 161-187", flatten_single_helix=True
        )
        return ha.run(start=10, stop=80)

    def test_regression_summary(self, helanal):
        bends = helanal.results.summary["all_bends"]
        old_helanal = read_bending_matrix(HELANAL_BENDING_MATRIX_SUBSET)
        assert_almost_equal(
            np.triu(bends["mean"], 1), old_helanal["Mean"], decimal=1
        )
        assert_almost_equal(
            np.triu(bends["sample_sd"], 1), old_helanal["SD"], decimal=1
        )
        assert_almost_equal(
            np.triu(bends["abs_dev"], 1), old_helanal["ABDEV"], decimal=1
        )

    def test_regression_values(self):
        u = mda.Universe(PDB_small)
        ha = hel.HELANAL(
            u, select="name CA and resnum 161-187", flatten_single_helix=True
        )
        ha.run()

        for key, value in HELANAL_SINGLE_DATA.items():
            if "summary" in key:
                data = ha.results[key.split()[0]]
                calculated = [
                    data.mean(),
                    data.std(ddof=1),
                    np.fabs(data - data.mean()).mean(),
                ]
            else:
                calculated = ha.results[key][0]

            msg = "Mismatch between calculated and reference {}"
            assert_almost_equal(
                calculated, value, decimal=4, err_msg=msg.format(key)
            )

    def test_multiple_selections(self, psf_ca):
        ha = hel.HELANAL(
            psf_ca,
            flatten_single_helix=True,
            select=("resnum 30-40", "resnum 60-80"),
        )
        ha.run()
        n_frames = len(psf_ca.universe.trajectory)
        assert len(ha.atomgroups) == 2
        assert len(ha.results.summary) == 2
        assert len(ha.results.all_bends) == 2
        assert ha.results.all_bends[0].shape == (n_frames, 8, 8)
        assert ha.results.all_bends[1].shape == (n_frames, 18, 18)

    def test_universe_from_origins(self, helanal):
        u = helanal.universe_from_origins()
        assert isinstance(u, mda.Universe)
        assert len(u.atoms) == len(helanal.atomgroups[0]) - 2
        assert len(u.trajectory) == 70

    def test_universe_from_origins_except(self, psf_ca):
        ha = hel.HELANAL(psf_ca, select="resnum 161-187")
        with pytest.raises(ValueError, match=r"before universe_from_origins"):
            u = ha.universe_from_origins()

    def test_multiple_atoms_per_residues(self):
        u = mda.Universe(XYZ)
        with pytest.warns(UserWarning) as rec:
            ha = hel.HELANAL(u, select="name H")
        assert len(rec) == 1
        assert "multiple atoms" in rec[0].message.args[0]

    def test_residue_gaps_split(self, psf_ca):
        sel = "resid 6:50 or resid 100:130 or resid 132:148"
        wmsg = "has gaps in the residues. Splitting into 3 helices"
        with pytest.warns(UserWarning, match=wmsg):
            ha = hel.HELANAL(psf_ca, select=sel).run()
            assert len(ha.atomgroups) == 3
            assert len(ha.atomgroups[0]) == 45
            assert len(ha.atomgroups[1]) == 31
            assert len(ha.atomgroups[2]) == 17

    def test_residue_gaps_no_split(self, psf_ca):
        sel = "resid 6:50 or resid 100:130 or resid 132:148"
        with pytest.warns(UserWarning) as rec:
            ha = hel.HELANAL(psf_ca, select=sel, split_residue_sequences=False)
            ha.run()
            assert len(ha.atomgroups) == 1
            assert len(ha.atomgroups[0]) == 45 + 31 + 17
        assert len(rec) == 1
        warnmsg = rec[0].message.args[0]
        assert "has gaps in the residues" in warnmsg
        assert "Splitting into" not in warnmsg

    def test_len_groups_short(self, psf_ca):
        sel = "resnum 161-168"
        with pytest.warns(UserWarning, match="Fewer than 9 atoms found"):
            ha = hel.HELANAL(psf_ca, select=sel)
            assert len(ha.atomgroups) < 9

    @pytest.mark.parametrize(
        "ref_axis,screw_angles",
        [
            # input vectors zigzag between [-1, 0, 0] and [1, 0, 0]
            # global axis is z-axis
            ([0, 0, 1], [180, 0]),  # calculated to x-z plane
            ([1, 0, 0], [180, 0]),  # calculated to x-z plane
            ([-1, 0, 0], [0, 180]),  # calculated to x-z plane upside down
            ([0, 1, 0], [90, -90]),  # calculated to y-z plane
            ([0, -1, 0], [-90, 90]),  # calculated to x-z plane upside down
            # calculated to diagonal xy-z plane rotating around
            ([1, 1, 0], [135, -45]),
            ([-1, 1, 0], [45, -135]),
            ([1, -1, 0], [-135, 45]),
            ([-1, -1, 0], [-45, 135]),
            # calculated to diagonal xyz-z plane w/o z contribution
            ([1, 1, 1], [135, -45]),
            ([1, 1, -1], [135, -45]),
            ([1, -1, 1], [-135, 45]),
            ([-1, 1, 1], [45, -135]),
            ([-1, -1, 1], [-45, 135]),
            ([-1, -1, -1], [-45, 135]),
        ],
    )
    def test_helanal_zigzag(self, zigzag, ref_axis, screw_angles):
        ha = hel.HELANAL(
            zigzag, select="all", ref_axis=ref_axis, flatten_single_helix=True
        ).run()
        assert_almost_equal(ha.results.local_twists, 180, decimal=4)
        assert_almost_equal(ha.results.local_nres_per_turn, 2, decimal=4)
        assert_almost_equal(ha.results.global_axis, [[0, 0, -1]], decimal=4)
        # all 0 vectors
        assert_almost_equal(ha.results.local_axes, 0, decimal=4)
        assert_almost_equal(ha.results.local_bends, 0, decimal=4)
        assert_almost_equal(ha.results.all_bends, 0, decimal=4)
        assert_almost_equal(ha.results.local_heights, 0, decimal=4)
        assert_almost_equal(
            ha.results.local_helix_directions[0][0::2],
            [[-1, 0, 0]] * 49,
            decimal=4,
        )
        assert_almost_equal(
            ha.results.local_helix_directions[0][1::2],
            [[1, 0, 0]] * 49,
            decimal=4,
        )
        origins = zigzag.atoms.positions[1:-1].copy()
        origins[:, 0] = 0
        assert_almost_equal(ha.results.local_origins[0], origins, decimal=4)
        assert_almost_equal(
            ha.results.local_screw_angles[0], screw_angles * 49, decimal=4
        )


def test_vector_of_best_fit():
    line = np.random.rand(3)
    unit = line / np.linalg.norm(line)
    points = line * np.arange(1000)[:, np.newaxis]
    noise = np.random.normal(size=(1000, 3))
    data = points + noise

    vector = hel.vector_of_best_fit(data)
    cos = np.dot(vector, unit)
    assert_almost_equal(abs(cos), 1.0, decimal=5)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

# Analyse a trajectory using elastic network models, following the approach of Hall et al (JACS 2007)
# Ben Hall (benjamin.a.hall@ucl.ac.uk) is to blame
# Copyright 2011; Consider under GPL v2 or later
r"""
Elastic network analysis of MD trajectories --- :mod:`MDAnalysis.analysis.gnm`
==============================================================================

:Author: Benjamin Hall <benjamin.a.hall@ucl.ac.uk>
:Year: 2011
:Copyright: Lesser GNU Public License v2.1 or later


Analyse a trajectory using elastic network models, following the approach of
:footcite:p:`Hall2007`.

An example is provided in the MDAnalysis Cookbook_, listed as GNMExample_.

.. _GNMExample: https://github.com/MDAnalysis/MDAnalysisCookbook/blob/master/examples/GNMExample.py
.. _Cookbook: https://github.com/MDAnalysis/MDAnalysisCookbook

The basic approach is to pass a trajectory to :class:`GNMAnalysis` and then run
the analysis:

.. code-block:: python

    u = MDAnalysis.Universe(PSF, DCD)
    C = MDAnalysis.analysis.gnm.GNMAnalysis(u, ReportVector="output.txt")

    C.run()
    output = zip(*C.results)

    with open("eigenvalues.dat", "w") as outputfile:
        for item in output[1]:
            outputfile.write(item + "\n")


The results are found in :attr:`GNMAnalysis.results`, which can be
used for further processing (see :footcite:p:`Hall2007`).

.. rubric:: References

.. footbibliography::


Analysis tasks
--------------

.. autoclass:: GNMAnalysis
   :members:
.. autoclass:: closeContactGNMAnalysis
   :members:

Utility functions
-----------------

The following functions are used internally and are typically not
directly needed to perform the analysis.

.. autofunction:: generate_grid
.. autofunction:: order_list

.. versionchanged:: 0.16.0
   removed unused function :func:`backup_file`

"""
import itertools
import logging
import warnings

import numpy as np

from .base import AnalysisBase, ResultsGroup


from MDAnalysis.analysis.base import Results

logger = logging.getLogger("MDAnalysis.analysis.GNM")


def _dsq(a, b):
    diff = a - b
    return np.dot(diff, diff)


def generate_grid(positions, cutoff):
    """Simple grid search.

    An alternative to searching the entire list of each atom; divide the
    structure into `cutoff` sized boxes This way, for each particle you only need
    to search the neighbouring boxes to find the particles within the `cutoff`.

    Observed a 6x speed up for a smallish protein with ~300 residues; this
    should get better with bigger systems.

    Parameters
    ----------
    positions : array
        coordinates of the atoms
    cutoff : float
        find particles with distance less than `cutoff` from each other; the
        grid will consist of boxes with sides of at least length `cutoff`

    """
    positions = np.asarray(positions)

    x, y, z = positions.T
    high_x = x.max()
    high_y = y.max()
    high_z = z.max()
    low_x = x.min()
    low_y = y.min()
    low_z = z.min()
    # Ok now generate a list with 3 dimensions representing boxes in x, y and z
    grid = [
        [
            [[] for i in range(int((high_z - low_z) / cutoff) + 1)]
            for j in range(int((high_y - low_y) / cutoff) + 1)
        ]
        for k in range(int((high_x - low_x) / cutoff) + 1)
    ]
    for i, pos in enumerate(positions):
        x_pos = int((pos[0] - low_x) / cutoff)
        y_pos = int((pos[1] - low_y) / cutoff)
        z_pos = int((pos[2] - low_z) / cutoff)
        grid[x_pos][y_pos][z_pos].append(i)
    return grid


def neighbour_generator(positions, cutoff):
    """
    return atom pairs that are in neighboring regions of space from a verlet-grid

    Parameters
    ----------
    positions : ndarray
        atom positions
    cutoff : float
        size of grid box

    Yields
    ------
    i_atom, j_atom
        indices of close atom pairs
    """
    grid = generate_grid(positions, cutoff)
    n_x = len(grid)
    n_y = len(grid[0])
    n_z = len(grid[0][0])
    for cell_x, cell_y, cell_z in itertools.product(
        range(n_x), range(n_y), range(n_z)
    ):
        atoms = grid[cell_x][cell_y][cell_z]
        # collect all atoms in own cell and neighboring cell
        all_atoms = []
        nei_cells = (-1, 0, 1)
        for x, y, z in itertools.product(nei_cells, nei_cells, nei_cells):
            gx = cell_x + x
            gy = cell_y + y
            gz = cell_z + z
            if 0 <= gx < n_x and 0 <= gy < n_y and 0 <= gz < n_z:
                all_atoms += grid[gx][gy][gz]
        # return all possible atom pairs in current cell
        for i_atom in atoms:
            for j_atom in all_atoms:
                yield i_atom, j_atom


def order_list(w):
    """Returns a dictionary showing the order of eigenvalues (which are reported scrambled normally)"""
    ordered = list(w)
    unordered = list(w)
    ordered.sort()
    list_map = {}
    for i in range(len(w)):
        list_map[i] = unordered.index(ordered[i])
    return list_map


class GNMAnalysis(AnalysisBase):
    """Basic tool for GNM analysis.

    Each frame is treated as a novel structure and the GNM
    calculated.  By default, this stores the dominant eigenvector
    and its associated eigenvalue; either can be used to monitor
    conformational change in a simulation.

    Parameters
    ----------
    universe : Universe
          Analyze the full trajectory in the universe.
    select : str (optional)
          MDAnalysis selection string
    cutoff : float (optional)
          Consider selected atoms within the cutoff as neighbors for the
          Gaussian network model.
    ReportVector : str (optional)
          filename to write eigenvectors to, by default no output is written
    Bonus_groups : tuple
          This is a tuple of selection strings that identify additional groups
          (such as ligands). The center of mass of each group will be added as
          a single point in the ENM (it is a popular way of treating small
          ligands such as drugs). You need to ensure that none of the atoms in
          `Bonus_groups` is contained in `selection` as this could lead to
          double counting. No checks are applied.

    Attributes
    ----------
    results.times : numpy.ndarray
            simulation times used in analysis
    results.eigenvalues : numpy.ndarray
            calculated eigenvalues
    results.eigenvectors : numpy.ndarray
            calculated eigenvectors

    See Also
    --------
    :class:`closeContactGNMAnalysis`


    .. versionchanged:: 0.16.0
       Made :meth:`generate_output` a private method :meth:`_generate_output`.

    .. versionchanged:: 1.0.0
       Changed `selection` keyword to `select`

    .. versionchanged:: 2.0.0
       Use :class:`~MDAnalysis.analysis.AnalysisBase` as parent class and
       store results as attributes ``times``, ``eigenvalues`` and
       ``eigenvectors`` of the ``results`` attribute.

    .. versionchanged:: 2.8.0
       Enabled **parallel execution** with the ``multiprocessing`` and ``dask``
       backends; use the new method :meth:`get_supported_backends` to see all
       supported backends.
    """

    _analysis_algorithm_is_parallelizable = True

    @classmethod
    def get_supported_backends(cls):
        return ("serial", "multiprocessing", "dask")

    def __init__(
        self,
        universe,
        select="protein and name CA",
        cutoff=7.0,
        ReportVector=None,
        Bonus_groups=None,
    ):
        super(GNMAnalysis, self).__init__(universe.trajectory)
        self.u = universe
        self.select = select
        self.cutoff = cutoff
        self.results = Results()
        self.results.eigenvalues = []
        self.results.eigenvectors = []
        self._timesteps = None  # time for each frame
        self.ReportVector = ReportVector
        self.Bonus_groups = (
            [self.u.select_atoms(item) for item in Bonus_groups]
            if Bonus_groups
            else []
        )
        self.ca = self.u.select_atoms(self.select)

    def _generate_output(
        self, w, v, outputobject, ReportVector=None, counter=0
    ):
        """Appends time, eigenvalues and eigenvectors to results.

        This generates the output by adding eigenvalue and
        eigenvector data to an appendable object and optionally
        printing some of the results to file. This is the function
        to replace if you want to generate a more complex set of
        outputs
        """
        list_map = order_list(w)
        if ReportVector:
            with open(ReportVector, "a") as oup:
                for item in enumerate(v[list_map[1]]):
                    print(
                        "",
                        counter,
                        item[0] + 1,
                        w[list_map[1]],
                        item[1],
                        file=oup,
                    )

        outputobject.eigenvalues.append(w[list_map[1]])
        outputobject.eigenvectors.append(v[list_map[1]])

    def generate_kirchoff(self):
        """Generate the Kirchhoff matrix of contacts.

        This generates the neighbour matrix by generating a grid of
        near-neighbours and then calculating which are are within
        the cutoff.

        Returns
        -------
        array
                the resulting Kirchhoff matrix
        """
        positions = self.ca.positions

        # add the com from each bonus group to the ca_positions list
        for item in self.Bonus_groups:
            # bonus = self.u.select_atoms(item)
            positions = np.vstack((positions, item.center_of_mass()))

        natoms = len(positions)
        matrix = np.zeros((natoms, natoms), np.float64)

        cutoffsq = self.cutoff**2

        for i_atom, j_atom in neighbour_generator(positions, self.cutoff):
            if (
                j_atom > i_atom
                and _dsq(positions[i_atom], positions[j_atom]) < cutoffsq
            ):
                matrix[i_atom][j_atom] = -1.0
                matrix[j_atom][i_atom] = -1.0
                matrix[i_atom][i_atom] = matrix[i_atom][i_atom] + 1
                matrix[j_atom][j_atom] = matrix[j_atom][j_atom] + 1

        return matrix

    def _single_frame(self):
        matrix = self.generate_kirchoff()
        try:
            _, w, v = np.linalg.svd(matrix)
        except np.linalg.LinAlgError:
            msg = f"SVD with cutoff {self.cutoff} failed to converge. "
            msg += f"Skip frame at {self._ts.time}."
            warnings.warn(msg)
            logger.warning(msg)
            return
        # Save the results somewhere useful in some useful format. Usefully.
        self._generate_output(
            w,
            v,
            self.results,
            ReportVector=self.ReportVector,
            counter=self._ts.frame,
        )

    def _conclude(self):
        self.results.times = self.times
        self.results.eigenvalues = np.asarray(self.results.eigenvalues)
        self.results.eigenvectors = np.asarray(self.results.eigenvectors)

    def _get_aggregator(self):
        return ResultsGroup(
            lookup={
                "eigenvectors": ResultsGroup.ndarray_hstack,
                "eigenvalues": ResultsGroup.ndarray_hstack,
                "times": ResultsGroup.ndarray_hstack,
            }
        )


class closeContactGNMAnalysis(GNMAnalysis):
    r"""GNMAnalysis only using close contacts.

    This is a version of the GNM where the Kirchoff matrix is
    constructed from the close contacts between individual atoms
    in different residues.

    Parameters
    ----------
    universe : Universe
          Analyze the full trajectory in the universe.
    select : str (optional)
          MDAnalysis selection string
    cutoff : float (optional)
          Consider selected atoms within the cutoff as neighbors for the
          Gaussian network model.
    ReportVector : str (optional)
          filename to write eigenvectors to, by default no output is written
    weights : {"size", None} (optional)
          If set to "size" (the default) then weight the contact by
          :math:`1/\sqrt{N_i N_j}` where :math:`N_i` and :math:`N_j` are the
          number of atoms in the residues :math:`i` and :math:`j` that contain
          the atoms that form a contact.

    Attributes
    ----------
    results.times : numpy.ndarray
            simulation times used in analysis
    results.eigenvalues : numpy.ndarray
            calculated eigenvalues
    results.eigenvectors : numpy.ndarray
            calculated eigenvectors

    Notes
    -----
    The `MassWeight` option has now been removed.

    See Also
    --------
    :class:`GNMAnalysis`


    .. versionchanged:: 0.16.0
       Made :meth:`generate_output` a private method :meth:`_generate_output`.

    .. deprecated:: 0.16.0
       Instead of ``MassWeight=True`` use ``weights="size"``.

    .. versionchanged:: 1.0.0
       MassWeight option (see above deprecation entry).
       Changed `selection` keyword to `select`

    .. versionchanged:: 2.0.0
       Use :class:`~MDAnalysis.analysis.AnalysisBase` as parent class and
       store results as attributes ``times``, ``eigenvalues`` and
       ``eigenvectors`` of the `results` attribute.
    """

    def __init__(
        self,
        universe,
        select="protein",
        cutoff=4.5,
        ReportVector=None,
        weights="size",
    ):
        super(closeContactGNMAnalysis, self).__init__(
            universe, select, cutoff, ReportVector
        )
        self.weights = weights

    def generate_kirchoff(self):
        nresidues = self.ca.n_residues
        positions = self.ca.positions
        residue_index_map = self.ca.resindices.copy()
        matrix = np.zeros((nresidues, nresidues), dtype=np.float64)
        cutoffsq = self.cutoff**2

        # cache sqrt of residue sizes (slow) so that sr[i]*sr[j] == sqrt(r[i]*r[j])
        inv_sqrt_res_sizes = np.ones(len(self.ca.residues))
        if self.weights == "size":
            inv_sqrt_res_sizes = 1 / np.sqrt(
                [r.atoms.n_atoms for r in self.ca.residues]
            )

        for i_atom, j_atom in neighbour_generator(positions, self.cutoff):
            if (
                j_atom > i_atom
                and _dsq(positions[i_atom], positions[j_atom]) < cutoffsq
            ):
                iresidue = residue_index_map[i_atom]
                jresidue = residue_index_map[j_atom]
                contact = (
                    inv_sqrt_res_sizes[iresidue] * inv_sqrt_res_sizes[jresidue]
                )
                matrix[iresidue][jresidue] -= contact
                matrix[jresidue][iresidue] -= contact
                matrix[iresidue][iresidue] += contact
                matrix[jresidue][jresidue] += contact

        return matrix



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import MDAnalysis as mda
import numpy as np
import pytest
from MDAnalysisTests.datafiles import COORDINATES_XTC, COORDINATES_TOPOLOGY
from numpy.testing import assert_almost_equal, assert_equal


def test_get_bad_auxreader_format_raises_ValueError():
    # should raise a ValueError when no AuxReaders with match the specified format
    with pytest.raises(ValueError):
        mda.auxiliary.core.get_auxreader_for(format="bad-format")


class BaseAuxReference(object):
    ## assumes the reference auxiliary data has 5 steps, with three values
    ## for each step: i, 2*i and 2^i, where i is the step number.
    ## If a particular AuxReader is such that auxiliary data is read in a
    ## format other than np.array([i, 2*i, 2**i]), format_data() should be
    ## overwritten tp return the appropriate format

    def __init__(self):
        self.n_steps = 5
        self.dt = 1
        self.initial_time = 0
        self.name = "test"

        # reference description of the (basic) auxiliary reader. Will
        # have to add 'format' and 'auxdata' when creating the reference
        # for each particular reader
        self.description = {
            "dt": self.dt,
            "represent_ts_as": "closest",
            "initial_time": self.initial_time,
            "time_selector": None,
            "data_selector": None,
            "constant_dt": True,
            "cutoff": None,
            "auxname": self.name,
        }

        def reference_auxstep(i):
            # create a reference AuxStep for step i
            auxstep = mda.auxiliary.base.AuxStep(
                dt=self.dt, initial_time=self.initial_time
            )
            auxstep.step = i
            auxstep._data = self.format_data([i, 2 * i, 2**i])
            return auxstep

        self.auxsteps = [reference_auxstep(i) for i in range(self.n_steps)]

        ## testing __getitem__ with slice and list. Should allow us to iterate
        ## through the specified auxiliary steps...
        self.iter_list = [1, -2]
        self.iter_list_auxsteps = [self.auxsteps[1], self.auxsteps[3]]
        self.iter_slice = slice(None, None, 2)  # every second step
        self.iter_slice_auxsteps = [
            self.auxsteps[0],
            self.auxsteps[2],
            self.auxsteps[4],
        ]

        def reference_timestep(dt=1, offset=0):
            # return a trajectory timestep with specified dt, offset + move to
            # frame 1; for use in auxiliary reading of different timesteps
            ts = mda.coordinates.base.Timestep(0, dt=dt, time_offset=offset)
            ts.frame = 1
            return ts

        ## test reading a timestep with lower frequency. Auxiliary steps with
        ## times between [1ps, 3ps) will be assigned to this timestep, i.e.
        ## step 1 (1 ps) and step 2 (2 ps).
        self.lower_freq_ts = reference_timestep(dt=2, offset=0)
        # 'closest' representative value will match step 2
        self.lowf_closest_rep = self.format_data([2, 2 * 2, 2**2])
        # 'average' representative value
        self.lowf_average_rep = self.format_data([1.5, 3, 3])

        ## test reading a timestep with higher frequency. Auxiliart steps with
        ## times between [0.25ps, 0.75ps) will be assigned to this timestep, i.e.
        ## no auxiliary steps
        self.higher_freq_ts = reference_timestep(dt=0.5, offset=0)
        self.highf_rep = self.format_data([np.nan, np.nan, np.nan])

        ## test reading a timestep that is offset. Auxiliary steps with
        ## times between [0.75ps, 1.75ps) will be assigned to this timestep, i.e.
        ## step 1 (1 ps)
        self.offset_ts = reference_timestep(dt=1, offset=0.25)
        # 'closest' representative value will match step 1 data
        self.offset_closest_rep = self.format_data([1, 2 * 1, 2**1])

        ## testing cutoff for representative values
        self.cutoff = 0
        # for 'average': use low frequenct timestep, only step 2 within 0ps cutoff
        self.lowf_cutoff_average_rep = self.format_data([2, 2 * 2, 2**2])
        # for 'closest': use offset timestep; no timestep within 0ps cutoff
        self.offset_cutoff_closest_rep = self.format_data(
            [np.nan, np.nan, np.nan]
        )

        ## testing selection of time/data. Overload for each auxilairy format
        ## as appropraite.
        # default None behavior set here so won't get errors when time/data
        # selection not implemented.
        self.time_selector = None
        self.select_time_ref = np.arange(self.n_steps)
        self.data_selector = None
        self.select_data_ref = [
            self.format_data([2 * i, 2**i]) for i in range(self.n_steps)
        ]

    def format_data(self, data):
        ## overload if auxiliary reader will read data with a format different
        ## to e.g. np.array([0, 0, 1])
        return np.array(data)


class BaseAuxReaderTest(object):

    def test_raise_error_no_auxdata_provided(self, ref, ref_universe):
        with pytest.raises(ValueError, match="No input `auxdata`"):
            ref_universe.trajectory.add_auxiliary()

    def test_n_steps(self, ref, reader):
        assert len(reader) == ref.n_steps, "number of steps does not match"

    def test_dt(self, ref, reader):
        assert reader.dt == ref.dt, "dt does not match"

    def test_initial_time(self, ref, reader):
        assert (
            reader.initial_time == ref.initial_time
        ), "initial time does not match"

    def test_first_step(self, ref, reader):
        # on first loading we should start at step 0
        assert_auxstep_equal(reader.auxstep, ref.auxsteps[0])

    def test_next(self, ref, reader):
        # should take us to step 1
        next(reader)
        assert_auxstep_equal(reader.auxstep, ref.auxsteps[1])

    def test_rewind(self, ref, reader):
        # move to step 1...
        reader.next()
        # now rewind should read step 0
        reader.rewind()
        assert_auxstep_equal(reader.auxstep, ref.auxsteps[0])

    def test_move_to_step(self, ref, reader):
        # should take us to step 3
        reader[3]
        assert_auxstep_equal(reader.auxstep, ref.auxsteps[3])

    def test_last_step(self, ref, reader):
        # should take us to the last step
        reader[-1]
        assert_auxstep_equal(reader.auxstep, ref.auxsteps[-1])

    def test_next_past_last_step_raises_StopIteration(self, ref, reader):
        # should take us to the last step
        reader[-1]
        # if we try to move to next step from here, should raise StopIteration
        with pytest.raises(StopIteration):
            reader.next()

    def test_move_to_invalid_step_raises_IndexError(self, ref, reader):
        # last step is number n_steps -1 ; if we try move to step number
        # n_steps we should get a ValueError
        with pytest.raises(IndexError):
            reader[ref.n_steps]

    def test_invalid_step_to_time_raises_ValueError(self, reader):
        # last step is number n_steps-1; if we try to run step_to_time on
        # step n_steps we should get a ValueError
        with pytest.raises(ValueError):
            reader.step_to_time(reader.n_steps)

    def test_iter(self, ref, reader):
        for i, val in enumerate(reader):
            assert_auxstep_equal(val, ref.auxsteps[i])

    def test_iter_list(self, ref, reader):
        # test using __getitem__ with a list
        for i, val in enumerate(reader[ref.iter_list]):
            assert_auxstep_equal(val, ref.iter_list_auxsteps[i])

    def test_iter_slice(self, ref, reader):
        # test using __getitem__ with a slice
        for i, val in enumerate(reader[ref.iter_slice]):
            assert_auxstep_equal(val, ref.iter_slice_auxsteps[i])

    def test_slice_start_after_stop_raises_IndexError(self, reader):
        # should raise IndexError if start frame after end frame
        with pytest.raises(IndexError):
            reader[2:1]

    def test_slice_out_of_range_raises_IndexError(self, ref, reader):
        # should raise IndexError if indices our of range
        with pytest.raises(IndexError):
            reader[ref.n_steps :]

    def test_slice_non_int_raises_TypeError(self, reader):
        # should raise TypeError if try pass in non-integer to slice
        with pytest.raises(TypeError):
            reader["a":]

    def test_bad_represent_raises_ValueError(self, reader):
        # if we try to set represent_ts_as to something not listed as a
        # valid option, we should get a ValueError
        with pytest.raises(ValueError):
            reader.represent_ts_as = "invalid-option"

    def test_time_selector(self, ref):
        # reload the reader, passing a time selector
        reader = ref.reader(ref.testdata, time_selector=ref.time_selector)
        # time should still match reference time for each step
        for i, val in enumerate(reader):
            assert (
                val.time == ref.select_time_ref[i]
            ), "time for step {} does not match".format(i)

    def test_time_non_constant_dt(self, reader):
        reader.constant_dt = False
        with pytest.raises(
            ValueError,
            match="If dt is not constant, must have a valid time selector",
        ):
            reader.time

    def test_time_selector_manual(self, ref):
        reader = ref.reader(ref.testdata, time_selector=ref.time_selector)
        # Manually set time selector
        reader.time_selector = ref.time_selector
        for i, val in enumerate(reader):
            assert (
                val.time == ref.select_time_ref[i]
            ), "time for step {} does not match".format(i)

    def test_data_selector(self, ref):
        # reload reader, passing in a data selector
        reader = ref.reader(ref.testdata, data_selector=ref.data_selector)
        # data should match reference data for each step
        for i, val in enumerate(reader):
            assert_equal(
                val.data,
                ref.select_data_ref[i],
                "data for step {0} does not match".format(i),
            )

    def test_no_constant_dt(self, ref):
        ## assume we can select time...
        # reload reader, without assuming constant dt
        reader = ref.reader(
            ref.testdata, time_selector=ref.time_selector, constant_dt=False
        )
        # time should match reference for selecting time, for each step
        for i, val in enumerate(reader):
            assert (
                val.time == ref.select_time_ref[i]
            ), "data for step {} does not match".format(i)

    def test_update_ts_without_auxname_raises_ValueError(self, ref):
        # reload reader without auxname
        with pytest.raises(ValueError):
            reader = ref.reader(ref.testdata)
            ts = ref.lower_freq_ts
            reader.update_ts(ts)

    def test_read_lower_freq_timestep(self, ref, reader):
        # test reading a timestep with lower frequency
        ts = ref.lower_freq_ts
        reader.update_ts(ts)
        # check the value set in ts is as we expect
        assert_almost_equal(
            ts.aux.test,
            ref.lowf_closest_rep,
            err_msg="Representative value in ts.aux does not match",
        )

    def test_represent_as_average(self, ref, reader):
        # test the 'average' option for 'represent_ts_as'
        # reset the represent method to 'average'...
        reader.represent_ts_as = "average"
        # read timestep; use the low freq timestep
        ts = ref.lower_freq_ts
        reader.update_ts(ts)
        # check the representative value set in ts is as expected
        assert_almost_equal(
            ts.aux.test,
            ref.lowf_average_rep,
            err_msg="Representative value does not match when "
            "using with option 'average'",
        )

    def test_represent_as_average_with_cutoff(self, ref, reader):
        # test the 'represent_ts_as' 'average' option when we have a cutoff set
        # set the cutoff...
        reader.cutoff = ref.cutoff
        # read timestep; use the low frequency timestep
        ts = ref.lower_freq_ts
        reader.update_ts(ts)
        # check representative value set in ts is as expected
        assert_almost_equal(
            ts.aux.test,
            ref.lowf_cutoff_average_rep,
            err_msg="Representative value does not match when "
            "applying cutoff",
        )

    def test_read_offset_timestep(self, ref, reader):
        # try reading a timestep offset from auxiliary
        ts = ref.offset_ts
        reader.update_ts(ts)
        assert_almost_equal(
            ts.aux.test,
            ref.offset_closest_rep,
            err_msg="Representative value in ts.aux does not match",
        )

    def test_represent_as_closest_with_cutoff(self, ref, reader):
        # test the 'represent_ts_as' 'closest' option when we have a cutoff set
        # set the cutoff...
        reader.cutoff = ref.cutoff
        # read timestep; use the offset timestep
        ts = ref.offset_ts
        reader.update_ts(ts)
        # check representative value set in ts is as expected
        assert_almost_equal(
            ts.aux.test,
            ref.offset_cutoff_closest_rep,
            err_msg="Representative value does not match when "
            "applying cutoff",
        )

    def test_read_higher_freq_timestep(self, ref, reader):
        # try reading a timestep with higher frequency
        ts = ref.higher_freq_ts
        reader.update_ts(ts)
        assert_almost_equal(
            ts.aux.test,
            ref.highf_rep,
            err_msg="Representative value in ts.aux does not match",
        )

    def test_get_auxreader_for(self, ref, reader):
        # check guesser gives us right reader
        reader = mda.auxiliary.core.get_auxreader_for(ref.testdata)
        assert reader == ref.reader

    def test_iterate_through_trajectory(self, ref, ref_universe):
        # check the representative values of aux for each frame are as expected
        # trajectory here has same dt, offset; so there's a direct correspondence
        # between frames and steps
        for i, ts in enumerate(ref_universe.trajectory):
            assert_equal(
                ts.aux.test,
                ref.auxsteps[i].data,
                "representative value does not match when "
                "iterating through all trajectory timesteps",
            )

    def test_iterate_as_auxiliary_from_trajectory(self, ref, ref_universe):
        # check representative values of aux for each frame are as expected
        # trajectory here has same dt, offset, so there's a direct correspondence
        # between frames and steps, and iter_as_aux will run through all frames
        for i, ts in enumerate(ref_universe.trajectory.iter_as_aux("test")):
            assert_equal(
                ts.aux.test,
                ref.auxsteps[i].data,
                "representative value does not match when "
                "iterating through all trajectory timesteps",
            )

    def test_auxiliary_read_ts_rewind(self, ref_universe):
        # AuxiliaryBase.read_ts() should retrieve the correct step after
        # reading the last one. Issue #2674 describes a case in which the
        # object gets stuck on the last frame.
        aux_info_0 = ref_universe.trajectory[0].aux.test
        ref_universe.trajectory[-1]
        aux_info_0_rewind = ref_universe.trajectory[0].aux.test
        assert_equal(
            aux_info_0,
            aux_info_0_rewind,
            "aux info was retrieved incorrectly "
            "after reading the last step",
        )

    def test_get_description(self, ref, reader):
        description = reader.get_description()
        for attr in ref.description:
            assert (
                description[attr] == ref.description[attr]
            ), "'Description' does not match for {}".format(attr)

    def test_load_from_description(self, reader):
        description = reader.get_description()
        new = mda.auxiliary.core.auxreader(**description)
        assert (
            new == reader
        ), "AuxReader reloaded from description does not match"

    def test_step_to_frame_out_of_bounds(self, reader, ref):

        ts = mda.coordinates.base.Timestep(0, dt=ref.dt)

        assert reader.step_to_frame(-1, ts) is None
        assert reader.step_to_frame(reader.n_steps, ts) is None

    def test_step_to_frame_no_time_diff(self, reader, ref):

        ts = mda.coordinates.base.Timestep(0, dt=ref.dt)

        for idx in range(reader.n_steps):

            assert reader.step_to_frame(idx, ts) == idx

    def test_step_to_frame_time_diff(self, reader, ref):

        # Timestep is 0.1 longer than auxiliary data
        ts = mda.coordinates.base.Timestep(0, dt=ref.dt + 0.1)

        # Test all 5 frames
        for idx in range(5):

            frame, time_diff = reader.step_to_frame(
                idx, ts, return_time_diff=True
            )

            assert frame == idx
            np.testing.assert_almost_equal(time_diff, idx * 0.1)

    def test_go_to_step_fail(self, reader):

        with pytest.raises(
            ValueError, match="Step index [0-9]* is not valid for auxiliary"
        ):
            reader._go_to_step(reader.n_steps)

    @pytest.mark.parametrize("constant", [True, False])
    def test_set_constant_dt(self, reader, constant):

        reader.constant_dt = constant

        assert reader.constant_dt == constant

    def test_copy(self, reader):
        new_reader = reader.copy()
        assert reader == new_reader
        assert reader is not new_reader


def assert_auxstep_equal(A, B):
    if not isinstance(A, mda.auxiliary.base.AuxStep):
        raise AssertionError("A is not of type AuxStep")
    if not isinstance(B, mda.auxiliary.base.AuxStep):
        raise AssertionError("B is not of type AuxStep")
    if A.step != B.step:
        raise AssertionError(
            "A and B refer to different steps: A.step = {}, "
            "B.step = {}".format(A.step, B.step)
        )
    if A.time != B.time:
        raise AssertionError(
            "A and B have different times: A.time = {}, "
            "B.time = {}".format(A.time, B.time)
        )
    if isinstance(A.data, dict):
        for term in A.data:
            assert_almost_equal(A.data[term], B.data[term])
    else:
        if any(A.data != B.data):
            # e.g. XVGReader
            raise AssertionError(
                "A and B have different data: A.data = {}, "
                "B.data = {}".format(A.data, B.data)
            )
